if __name__ != "__dpl__":
    raise Exception

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

ext = dpl.extension(meta_name="tests", alias=__alias__)

ext.items["ASSERT_ERROR"] =\
dpl.register_error("ASSERT_ERROR")

def pprint(d, l=0, seen=None):
    if seen is None:
        seen = set()
    if id(d) in seen:
        print("  "*l+"...")
        return
    seen.add(id(d))
    if isinstance(d, list):
        for i in d:
            if isinstance(i, list):
                print("  "*l+f"[")
                pprint(i, l+1, seen)
                print("  "*l+"]")
            else:
                print("  "*l+repr(i))
        return
    for name, value in d.items():
        if name.startswith("_"):
            ...
        elif isinstance(value, dict):
            print("  "*l+"{name!r} => {{")
            pprint(value, l+1, seen)
            print("  "*l+"}")
        elif isinstance(value, list):
            print("  "*l+f"{name!r} => [")
            pprint(value, l+1, seen)
            print("  "*l+"]")
        else:
            print(f"{name!r} = {value!r}")

@ext.add_func("assert", "$$ :: any\n$$[2] :: any str")
def _(frame, _, condition, message='Not provided.'):
    "Raises an error"
    if not condition:
        return f"err:{ext.items['ASSERT_ERROR']}:{message}"


@ext.add_func(typed="$$ :: any\n$$[2] :: any str")
def quiet_assert(frame, _, condition, message='Not provided'):
    "Does not raise an error"
    if not condition:
        print("Assert Failed:", message, flush=True)



from io import TextIOWrapper

if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

ext = dpl.extension(meta_name="io", alias=__alias__)

dpl.type_checker.alias["TextIOWrapper"] = TextIOWrapper

@ext.add_func("open", "%$$[str,str]\n%$$[str]")
def myOpen(_, local, file_name, mode="r"):
    try:
        if modules.os.path.isabs(file_name):
            file = file_name
        else:
            file = modules.os.path.join(local, file_name)
        return (open(file),)
    except Exception as e:
        return dpl.error.get_error_string("PYTHON_ERROR", repr(e))


@ext.add_func(typed="$$[4] :: str \"=\" str str \n $$[3] :: str \"=\" str")
def with_file(frame, local, body, name, _eq_, file_name, mode="r"):
    if _eq_ != "=":
        return dpl.error.get_error_string("SYNTAX_ERROR", f"Expected it to be `body io:with_file {name} **=** {file_name!r} {mode!r}`") 
    try:
        if modules.os.path.isabs(file_name):
            file = file_name
        else:
            file = modules.os.path.join(local, file_name)
        with open(file, mode) as f:
            frame[-1][name] = f
            return dpl.run_code(body, frame=frame)
    except Exception as e:
        return dpl.error.get_error_string("PYTHON_ERROR", repr(e))

@ext.add_func(typed="$$[2] :: TextIOWrapper int \n $$[3] :: TextIOWrapper int int")
def seek(_, __, file_object, position, whence=0):
    file_object.seek(position, whence)


@ext.add_func(typed="$$ :: TextIOWrapper")
def read(_, __, file_object):
    try:
        return (file_object.read(),)
    except Exception as e:
        file_object.close()
        return (e,)


@ext.add_func(typed="$$ :: TextIOWrapper str")
def write(_, __, file_object, content):
    try:
        file_object.write(content)
    except Exception as e:
        file_object.close()
        return f"err:{dpl.error.PYTHON_ERROR}:{repr(e)}"


@ext.add_func(typed="$$ :: TextIOWrapper str")
def append(_, __, file_object, content):
    try:
        if (mode := file_object.mode) == "a":
            file_object.append(content)
        else:
            raise Exception(
                f'Invalid operation on a file! Expected the mode to be "a" but got "{mode}"'
            )
    except Exception as e:
        file_object.close()
        return f"err:{dpl.error.PYTHON_ERROR}:{repr(e)}"


@ext.add_func(typed="$$ :: TextIOWrapper")
def close(_, __, file_object):
    file_object.close()



# Utilities for QOL functions
# Now includes DSL Config Parser

from . import arguments as argproc

def flatten_dict(d, parent_key="", sep=".", seen=None):
    if seen is None:
        seen = set()
    items = {}
    dict_id = id(d)
    if dict_id in seen:
        return d
    seen.add(dict_id)
    for key, value in d.items():
        new_key = f"{parent_key}{sep}{key}" if parent_key else key
        if isinstance(value, dict):
            items.update(flatten_dict(value, new_key, sep, seen))
        elif not isinstance(key, str):
            continue
        else:
            items[new_key] = value
    seen.remove(dict_id)
    return items

def convert_sec(sec):
    "Convert seconds to appropriate units"
    if sec >= 1:
        return sec, "s"
    elif sec >= 1e-3:
        return sec * 1e3, "ms"
    elif sec >= 1e-6:
        return sec * 1e6, "µs"
    elif sec >= 1e-9:
        return sec * 1e9, "ns"
    else:
        return sec * 1e12, "ps"


def convert_bytes(byte):
    "Convert bytes to appropriate units"
    if byte < 1e3:
        return byte, "B"
    elif byte < 1e6:
        return byte * 1e-3, "KB"
    elif byte < 1e9:
        return byte * 1e-6, "MB"
    elif byte < 1e12:
        return byte * 1e-9, "GB"
    elif byte < 1e15:
        return byte * 1e-12, "TB"
    else:
        return byte * 1e-15, "PB"

def format_bytes(byte):
    vv, vu = convert_bytes(byte)
    return f"{vv:.2f}{vu}"

def format_seconds(byte):
    vv, vu = convert_sec(byte)
    return f"{vv:.2f}{vu}"


# py_parser2.py - The brother of the parser.
# prototype for the new execution loop

from .runtime import *
from .py_parser import process, get_block
import traceback


def get_var(frame, name, default=constants.nil, resolve_name=False):
    "Helper function to automatically resolve to global scope."
    value = rget(frame[-1], name)
    if value == constants.nil:
        value = rget(frame[0], name, default=default)
        name = f"_global.{name}"
    if resolve_name:
        return name, value
    else:
        return value

class instruction:
    class resolved:
        def __init__(self, name, func):
            self.value = func
            self.name = name
        def __repr__(self):
            return self.name

class executor:
    "A class that can be used easily to run DPL code,\nand even your own language if modified."
    instructions = {}
    def __init__(self, code, frame=None):
        "Make an executor instance."
        if isinstance(code, str):
            code = process(code)
        if isinstance(code, int):
            raise error.DPLError(code) from None
        elif isinstance(code, list):
            self.code = code
            self.frame = frame or new_frame()
        else:
            self.code, self.frame = code["code"], code["frame"]
        if frame:
            self.frame[0].update(frame[0])
        self.instruction_pointer = 0
    @staticmethod
    def add_intrinsic(name=None):
        "Add an intrinsic.\nBuiltin functions have already been defined in executor.instructions,\nclear it to define your own."
        def wrapper(func):
            executor.instructions[name or func.__name__] = func
            return func
        return wrapper
    def resolve_instructions(self):
        "Resolve the instructions."
        cache = []
        for file, pos, ins, oargs in self.code:
            if ins in executor.instructions:
                cache.append((file, pos, instruction.resolved(ins, executor.instructions[ins]), oargs))
            else:
                cache.append((file, pos, ins, oargs))
        self.code.clear()
        self.code.extend(cache)
    def run(self):
        "Run the code inside the current instance."
        while self.instruction_pointer < len(self.code):
            file, pos, ins, oargs = self.code[self.instruction_pointer]
            if not oargs is None:
                try:
                    args = process_args(self.frame, oargs)
                    argc = len(args)
                    if debug_settings["type_checker"]:
                        if (tmp:=(pos, file, ins)) not in tc_cache:
                            tc_cache.add(tmp)
                            if not check_ins(ins, args):
                                itypesr = get_ins(ins, args)
                                if itypesr is None and not varproc.is_debug_enabled("TC_DEFAULT_WHEN_NOT_FOUND"):
                                    error.error(pos, file, f"Type signature for {ins} is not defined.\nUse tc_register to register your function.")
                                    return error.TYPE_ERROR
                                itypes = tuple(map(lambda x: getattr(x, "__name__", x), itypesr))
                                atypes = tuple(map(lambda x: type(x).__name__, args))
                                error.error(pos, file, f"Type mismatch [{ins}]: Expected {itypes} but got {atypes}")
                                return error.TYPE_ERROR
                except Exception as e:
                    error.error(
                        pos,
                        file,
                        f"Something went wrong when arguments were processed:\n{e}\n> {oargs!r}",
                    )
                    return error.PYTHON_ERROR
            else:
                args = []
                argc = 0
            # Instead of O(n) we can get O(1)
            # or at least the closest we can go.
            if not isinstance(ins, instruction.resolved):
                if (func:=get_var(self.frame, ins)) != constants.nil \
                and isinstance(func, dict):
                    nscope(self.frame)
                    self.frame[-1].update(dict(zip(func["args"], args)))
                    if func["self"]:
                        self.frame[-1]["self"] = func["self"]
                    if func["capture"]:
                        self.frame[-1]["_capture"] = func["capture"]
                    res = executor(func["body"], self.frame).run()
                    pscope(self.frame)
                    if res == error.STOP_FUNCTION:
                        self.instruction_pointer += 1
                        continue
                    if res: return res
                    self.instruction_pointer += 1
                    continue
                elif hasattr(func, "__code__"):
                    try:
                        func_params = ext_s.get_py_params(func)[2:]
                        if func_params and any(map(lambda x: x.endswith("_body") or x.endswith("_xbody"), func_params)):
                            t_args = []
                            for i in func_params:
                                if i.endswith("_xbody"):
                                    temp = get_block(code, instruction_pointer, start=0)
                                    if temp is None:
                                        error.error(pos, file, f"Function '{function.__name__}' expected a block!")
                                        return (error.RUNTIME_ERROR, error.SYNTAX_ERROR)
                                    instruction_pointer, body = temp
                                    t_args.append(body)
                                elif i.endswith("_body"):
                                    temp = get_block(code, instruction_pointer)
                                    if temp is None:
                                        error.error(pos, file, f"Function '{function.__name__}' expected a block!")
                                        return (error.RUNTIME_ERROR, error.SYNTAX_ERROR) # say "runtime error" caised by "syntax error"
                                    instruction_pointer, body = temp
                                    t_args.append(body)
                                else:
                                    t_args.append(args.pop(0))
                        else:
                            t_args = args
                        res = ext_s.call(
                            func, self.frame, varproc.meta_attributes["internal"]["main_path"], t_args
                        )
                        if isinstance(res, int) and res:
                            return res
                        elif isinstance(res, str):
                            if res == "break":
                                break
                            elif res.startswith("err:"):
                                _, ecode, message = res.split(":", 2)
                                error.error(pos, file, message)
                                return int(ecode)
                    except:
                        error.error(pos, file, traceback.format_exc()[:-1])
                        return error.PYTHON_ERROR
                    self.instruction_pointer += 1
                    continue
                else:
                    error.error(
                        pos, file,
                        f"Invalid instruction: {ins}"
                    )
                    return error.RUNTIME_ERROR
            try:
                if err:=ins.value(self, file, pos, *args):
                    if err == error.STOP_FUNCTION:
                        self.instruction_pointer += 1
                        continue
                    if err: return err
            except Exception as e:
                error.error(
                    pos, file,
                    traceback.format_exc()
                )
                return error.PYTHON_ERROR
            self.instruction_pointer += 1
        return 0
    def reset(self):
        "Reset the current instance."
        self.instruction_pointer = 0

# implement bare minimum for testing.

@executor.add_intrinsic()
def _test(*_):
    print("This was a test for the new parser.", _)

@executor.add_intrinsic("_intern.switch")
def _(parser, file, pos, switch_body, switch_value):
    return executor(switch_body.get(switch_value, switch_body[None]), frame=parser.frame).run()

@executor.add_intrinsic("set")
def _(parser, file, pos, name, _eq, value):
    if _eq != "=":
        return error.SYNTAX_ERROR
    rset(parser.frame[-1], name, value)

@executor.add_intrinsic("if")
def _(parser, file, pos, value):
    temp = get_block(parser.code, parser.instruction_pointer)
    if not temp:
        return error.SYNTAX_ERROR
    parser.instruction_pointer, body = temp
    if value and body:
        return executor(body, frame=parser.frame).run()

@executor.add_intrinsic()
def fn(parser, file, pos, name, params):
    temp = get_block(parser.code, parser.instruction_pointer)
    if not temp:
        return error.SYNTAX_ERROR
    parser.instruction_pointer, body = temp
    rset(parser.frame[-1], name, make_function(
        name = name,
        body = body,
        params = params
    ))

@executor.add_intrinsic()
def loop(parser, file, pos, iterations=None):
    temp = get_block(parser.code, parser.instruction_pointer)
    if not temp:
        return error.SYNTAX_ERROR
    parser.instruction_pointer, body = temp
    if not body:
        return
    if iterations is None:
        while True:
            err = executor(body, frame=parser.frame).run()
            if err == error.SKIP_RESULT:
                continue
            elif err == error.STOP_RESULT or err == error.STOP_FUNCTION:
                break
            elif err > 0:
                break
    else:
        for _ in range(iterations):
            err = executor(body, frame=parser.frame).run()
            if err == error.SKIP_RESULT:
                continue
            elif err == error.STOP_RESULT or err == error.STOP_FUNCTION:
                break
            elif err > 0:
                break

@executor.add_intrinsic()
def inc(parser, file, pos, name):
    resolved_name, value = get_var(parser.frame, name, default=0, resolve_name=True)
    rset(parser.frame[-1], resolved_name, value+1)

@executor.add_intrinsic()
def dec(parser, file, pos, name):
    resolved_name, value = get_var(parser.frame, name, default=0, resolve_name=True)
    rset(parser.frame[-1], resolved_name, value-1)

@executor.add_intrinsic()
def START_TIME(parser, file, pos):
    parser.start_time = time.perf_counter()

@executor.add_intrinsic()
def STOP_TIME(parser, file, pos):
    parser.end_time = time.perf_counter()

@executor.add_intrinsic()
def LOG_TIME(parser, file, pos, message=None):
    ct, unit = utils.convert_sec(parser.end_time - parser.start_time)
    error.info(f"Elapsed time: {(' '+message+' ') if message else ''} {ct:,.8f}{unit}")

@executor.add_intrinsic()
def match(parser, file, pos, value):
    temp = get_block(parser.code, parser.instruction_pointer)
    if temp is None:
       return error.SYNTAX_ERROR
    else:
        parser.instruction_pointer, body = temp
        if (err := parse_match(parser.frame, body, value)) > 0:
           return err

@executor.add_intrinsic("println")
def _(_, __, ___, *values):
    print(*values)

@executor.add_intrinsic("print")
def _(_, __, ___, *values):
    print(*values, end="")

@executor.add_intrinsic("while")
def _(parser, file, pos, value):
    temp = get_block(parser.code, parser.instruction_pointer)
    if temp is None:
       return error.SYNTAX_ERROR
    else:
        parser.instruction_pointer, body = temp
        print(body)
        if not body:
            return
        while evaluate(parser.frame, value):
            err = executor(body, frame=parser.frame).run()
            if err == error.SKIP_RESULT:
                continue
            elif err == error.STOP_RESULT or err == error.STOP_FUNCTION:
                break
            elif err > 0:
                break

@executor.add_intrinsic("for")
def _(parser, file, pos, name, _in, value):
    if _in != "in":
        return error.SYNTAX_ERROR
    temp = get_block(parser.code, parser.instruction_pointer)
    if temp is None:
       return error.SYNTAX_ERROR
    else:
        parser.instruction_pointer, body = temp
        if not body:
            return
        for i in value:
            parser.frame[-1][name] = i
            err = executor(body, frame=parser.frame).run()
            if err == error.SKIP_RESULT:
                continue
            elif err == error.STOP_RESULT or err == error.STOP_FUNCTION:
                break
            elif err > 0:
                break

@executor.add_intrinsic()
def catch(parser, file, pos, values, _eq, func_name, *args):
    if _eq != "=":
        return error.SYNTAX_ERROR
    func = get_var(parser.frame, func_name)
    if func == constants.nil:
        error.error(pos, file, f"{func_name} does not exist!")
        return error.NAME_ERROR
    nscope(parser.frame)
    parser.frame[-1]["_returns"] = values
    parser.frame[-1].update(dict(zip(func["args"], args)))
    if func["self"]:
        parser.frame[-1]["self"] = func["self"]
    if func["capture"]:
        parser.frame[-1]["_capture"] = func["capture"]
    res = executor(func["body"], parser.frame).run()
    pscope(parser.frame)
    if res == error.STOP_FUNCTION:
        ...
    elif res: return res

@executor.add_intrinsic("return")
def _(parser, file, pos, *values):
    if (temp:=parser.frame[-1].get("_returns")) is not None:
        for name, value in zip(temp, values):
            rset(parser.frame[-1], f"_nonlocal.{name}", value)
    return error.STOP_FUNCTION

@executor.add_intrinsic()
def skip(_, __, ___):
    return error.SKIP_RESULT

@executor.add_intrinsic()
def stop(_, __, ___):
    return error.STOP_RESULT

@executor.add_intrinsic("exit")
def _(_, __, ___, code=0):
    exit(code)

@executor.add_intrinsic("_intern.jump")
def _(parser, _, __, index, condition=...):
    if condition == ...:
        parser.instruction_pointer = index
    else:
        if condition:
            parser.instruction_pointer = index

@executor.add_intrinsic("_intern.get_index")
def _(parser, _, __, name):
    rset(parser.frame[-1], name, parser.instruction_pointer)


# A module to help the parser call dynamically loaded functions and to load them too
# This is needed for the cythonized parser and slows down the pure python impl!
# I tried it, it still failed :)

import itertools
import time
import os, sys
import traceback
from . import type_checker
from . import utils
from . import varproc
from . import arguments as argproc
from . import info
if "no-lupa" not in info.program_flags:
    import lupa
from types import ModuleType
from . import error
from . import state
from . import restricted
from . import objects
from . import constants
from . import py_argument_handler
arguments_handler = py_argument_handler.arguments_handler


def register_run(func):
    dpl.run_code = func


def register_process(func):
    dpl.process_code = func


class modules:
    "Capsule for modules."

    os = os
    sys = sys
    traceback = traceback
    time = time
    itertools = itertools


class extension:
    "A class to help define methods and functions."

    def __init__(self, name=None, meta_name=None, alias=None):
        self.__func = {}  # functions
        self.__meth = {}  # methods
        self.__data = {}
        self._name = (
            name  # This is a scope name, dpl defined name.func_name
        )
        self.meta_name = meta_name  # while this is the mangled name, python defined "{meta_name}:{func_name}"
        if not alias is None:
            if self._name:
                self._name = alias
            elif self.meta_name:
                self.meta_name = alias
            else:
                self._name = alias

    def add_func(self, name=None, typed=None):
        "Add a function."
        def wrap(func):
            nonlocal name
            if func.__doc__ is None:
                func.__doc__ = (
                    f"Function `{self.meta_name}:{name}`"
                    if self.meta_name
                    else f"Function {self._name}.{name}"
                ) + ": Default doc string..."
            if name is None:
                name = getattr(func, "__name__", None) or "_"
            self.__func[self.mangle(name)] = (
                func
            )
            if typed:
                type_checker.register(typed.replace("$$", self.mangle(name)))
            return func
        return wrap

    def add_method(self, name=None, from_func=False):
        "Add a method."
        def wrap(func):
            nonlocal name
            if name is None:
                name = getattr(func, "__name__", None) or "_"
            self.__meth[self.mangle(name)] = (
                func if not from_func else lambda *args: func(args[0], None, *args[1:])
            )
            return func
        return wrap

    def get(self, name, default=None):
        return self.__data.get(name, default)

    def mangle(self, name):
        return (
            f"{self._name}.{name}"
            if not self.meta_name
            else f"{self.meta_name}:{name}"
        )

    @property
    def name(self):
        return (self._name
            if not self.meta_name
            else self.meta_name)

    @property
    def functions(self):
        return self.__func

    @property
    def methods(self):
        return self.__meth

    @property
    def items(self):
        return self.__func

    def __repr__(self):
        return f"Extension<{self._name or f'{self.meta_name}:*'!r}>"


def require(path):
    "Import a python file in the lib dir.\nIn cases of 'dir/.../file' use ['dir', ..., 'file'],\nthis uses os.path.join to increase portability."
    mod = {
        "__name__": "__dpl_require__",
        "modules": modules,
        "dpl": dpl,
        "__import__": restricted.restricted(__import__),
    }
    if isinstance(path, (list, tuple)):
        path = os.path.join(*path)
    path = os.path.join(info.LIBDIR, path)
    if not os.path.isfile(path):
        raise FileNotFoundError(f"{path!r} not found.")
    try:
        with open(path, "r") as f:
            exec(compile(f.read(), path, "exec"), mod)
        r = ModuleType(path)
        for name, value in mod.items():
            setattr(r, name, value)
        return r
    except Exception as e:
        return e


class wrap:
    def __init__(self, data):
        self.type = type(data) if not data is None else None
        self.value = data
    def __repr__(self):
        return f"wrap(type={self.type}, value={self.value!r})"

class dpl:
    require = require
    utils = utils
    varproc = varproc
    arguments = argproc
    info = info
    error = error
    state = state
    ffi = None
    register_error = error.register_error
    restricted = restricted
    state_nil = state.bstate("nil")
    state_none = state.bstate("none")
    constants = constants
    state_true = 1
    state_false = 0
    extension = extension
    objects = objects
    falsy = (state_nil, state_none, state_false, None, False)
    truthy = (state_true, True)
    exit = None
    wrap = wrap

    def pycall(func, args, table):
        return func(*(args or []), **table)
    
    def add_matcher(name):
        def wrap(fn):
            argproc.matches[name] = fn
            return fn
        return wrap

    type_checker = type_checker


def get_py_params(func):
    if not hasattr(func, "__code__"):
        return None
    co = func.__code__
    arg_count = co.co_argcount
    return co.co_varnames[:arg_count]

def luaj_import(
    frame, file, search_path=None, loc=varproc.meta_attributes["internal"]["main_path"]
):
    lua = lupa.LuaRuntime(unpack_returned_tuples=True)
    if not os.path.isabs(file):
        if search_path is not None:
            file = os.path.join(
                {"_std": info.LINDIR, "_loc": loc}.get(
                    search_path, search_path
                ),
                file,
            )
        if not os.path.isfile(file):
            print("File not found:", file)
            return 1
    if os.path.isdir(file):
        if os.path.isfile(files:=os.path.join(file, "include-lua.txt")):
            with open(files) as f:
                for line in f:
                    line = line.strip()
                    if "=>" in line:
                        line, alias = line.split("=>")
                        line = line.strip()
                        alias = alias.strip()
                    else:
                        alias = None
                    if line.startswith("#:"):
                        print(f"{files} [N/A]:",line[2:]) # for messages like deprecation warnings
                    elif line.startswith("#?"):
                        print(line[2:]) # for messages like deprecation warnings
                    elif line.startswith('#') or not line:
                        ...
                    else:
                        if luaj_import(frame, line, search_path=file, loc=loc):
                            print(f"Something went wrong while importing {line!r}")
                            return 1
            return
        else:
            print("luaj: 'include-lua.txt' not found.\nTried to include a directory ({file!r}) without an include file!")
            return 1
    if varproc.is_debug_enabled("show_imports"):
        error.info(f"Imported {file!r}")
    with open(file, "r") as f:
        try:
            lua.globals()["__dpl__"] = info.VERSION
            lua.globals()["package"]["path"] = (
                info.LIBDIR + os.sep + "?.lua;" + lua.globals()["package"]["path"]
            )
            lua.globals()["api"] = {
                "dpl": dpl,
                "modules": modules,
                "types": {"tuple": tuple, "set": set, "list": list},
            }
            lua.execute(f.read(), file)
        except (SystemExit, KeyboardInterrupt):
            raise
        except:
            error.error("[N/A]", file, traceback.format_exc())
            return 1
    funcs = {}
    meths = {}
    for name in lua.globals():
        ext = lua.globals()[name]
        if isinstance(ext, extension):
            if ext.name in frame[-1]:
                raise Exception(f"Name clashing! For name {ext.name!r}")
            if ext.name:
                varproc.rset(frame[-1], ext.name, (temp := {}))
                temp.update(ext.functions)
            else:
                funcs.update(ext.functions)
            meths.update(ext.methods)
    frame[-1].update(funcs)
    argproc.methods.update(meths)
    file = os.path.realpath(file)
    if search_path in varproc.dependencies["lua"]:
        varproc.dependencies["lua"][search_path].add(file)
    else:
        varproc.dependencies["lua"][search_path] = {file}


def py_import(frame, file, search_path=None, loc=varproc.meta_attributes["internal"]["main_path"], alias=None):
    if not os.path.isabs(file):
        if search_path is not None:
            file = os.path.join(
                {"_std": info.LIBDIR, "_loc": loc}.get(
                    search_path, search_path
                ),
                file,
            )
        if not os.path.exists(file):
            print("Not found:", file)
            return 1
    if os.path.isdir(file):
        if alias:
            frame[-1][alias] = {}
            frame = [frame[-1][alias]]
        if os.path.isfile(files:=os.path.join(file, "include-py.txt")):
            with open(files) as f:
                for line in f.readlines():
                    line = line.strip()
                    if "=>" in line:
                        line, alias = line.split("=>")
                        line = line.strip()
                        alias = alias.strip()
                    else:
                        alias = None
                    if line.startswith("#:"):
                        print(f"{files} [N/A]:",line[2:]) # for messages like deprecation warnings
                    elif line.startswith("#?"):
                        print(line[2:]) # for messages like deprecation warnings
                    elif line.startswith('#') or not line:
                        ...
                    else:
                        if py_import(frame, line, search_path=file, loc=loc):
                            print(f"Something went wrong while importing {line!r}")
                            return 1
            return
        else:
            print(f"python: 'include-py.txt' not found.\nTried to include a directory ({file!r}) without an include file!")
            return 1
    if varproc.is_debug_enabled("show_imports"):
        error.info(f"Imported {file!r}" if not alias else f"Imported {file!r} as {alias}")
    with open(file, "r") as f:
        obj = compile(f.read(), file, "exec")
        try:
            d = {"__name__": "__dpl__", "modules": modules, "dpl": dpl, "__alias__":alias}
            exec(obj, d)
        except (SystemExit, KeyboardInterrupt):
            raise
        except:
            error.error("[N/A]", file, traceback.format_exc())
            return 1
    funcs = {}
    meths = {}
    for name, ext in d.items():
        if isinstance(ext, extension):
            if ext.meta_name:
                funcs.update(ext.functions)
            elif ext.name in frame[-1]:
                raise Exception(f"Name clashing! For name {ext.name!r}")
            elif ext.name:
                varproc.rset(frame[-1], ext.name, (temp := {}))
                temp.update(ext.functions)
            meths.update(ext.methods)
    frame[-1].update(funcs)
    argproc.methods.update(meths)
    file = os.path.realpath(file)
    if search_path in varproc.dependencies["python"]:
        varproc.dependencies["python"][search_path].add(file)
    else:
        varproc.dependencies["python"][search_path] = {file}


def call(func, frame, file, args):
    if args and isinstance(args[0], arguments_handler):
        if args[0].args:
            args[0].args.insert(0, file)
            args[0].args.insert(0, frame)
        else:
            args[0].args = [frame, file]
        ret = args[0].call(func)
    else:
        ret = func(frame, file, *args)
    return ret


from . import py_parser as parser
import os
import fnmatch
import ast
from collections import defaultdict

def convert_bytes(byte):
    "Convert bytes to appropriate units"
    if byte < 1e3:
        return byte, "B"
    elif byte < 1e6:
        return byte * 1e-3, "KB"
    elif byte < 1e9:
        return byte * 1e-6, "MB"
    elif byte < 1e12:
        return byte * 1e-9, "GB"
    elif byte < 1e15:
        return byte * 1e-12, "TB"
    else:
        return byte * 1e-15, "PB"

class FileWalker:
    def __init__(self):
        self.filetype_processors = {}
        self.deps = defaultdict(dict)

    def add_filetype(self, pattern, processor):
        self.filetype_processors[pattern] = processor

    def _match_processor(self, filename):
        for pattern, processor in self.filetype_processors.items():
            if fnmatch.fnmatch(filename, pattern):
                return processor
        return None

    def scan(self, root_path):
        for dirpath, _, filenames in os.walk(root_path):
            for filename in filenames:
                processor = self._match_processor(filename)
                if processor:
                    full_path = os.path.join(dirpath, filename)
                    try:
                        processor(self.deps, full_path)
                    except Exception as e:
                        print(f"[!] Failed to process {full_path}: {e}")

    def get_results(self):
        return dict(self.deps)

def py_processor(deps, path):
    with open(path, 'r', encoding='utf-8') as f:
        try:
            tree = ast.parse(f.read(), filename=path)
        except SyntaxError:
            return
    deps[path] = {
        "definitions":[],
        "file_size":os.path.getsize(path)
    }
    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            args = [arg.arg for arg in node.args.args]
            if node.args.vararg:
                args.append(f"*{node.args.vararg.arg}")
            if node.args.kwarg:
                args.append(f"**{node.args.kwarg.arg}")
            deps[path]["definitions"].append({
                "name": node.name,
                "args": args,
                "line": node.lineno,
                "doc": ast.get_docstring(node) or "[not found]"
            })

def py_processor(deps, path):
    with open(path, 'r', encoding='utf-8') as f:
        try:
            tree = ast.parse(f.read(), filename=path)
        except SyntaxError:
            return
    deps[path] = {
        "definitions":[],
        "file_size":os.path.getsize(path)
    }

if __name__ == "__main__":
    from pprint import pprint
    import sys
    if len(sys.argv) < 2:
        print("Usage: python script.py <directory>")
    else:
        walker = FileWalker()
        walker.add_filetype("*.py", py_processor)
        walker.scan(argv[1])
        for file, files in walker.get_results().items():
            print("File:", file)
            for file in files["definitions"]:
                print(f"""    Function: {file['name']}({', '.join(file['args'])})
        Doc-str:
        {file['doc'].replace("\n", "\n        ")}
        Line: {file['line']}""")
            uv, us = convert_bytes(files["file_size"])
            print(f"File size: {uv}{us}")


# Logging

import datetime
import os

class DPLError(Exception):
    def __init__(self, code):
        self._code = code
        self._name = ERRORS_DICT.get(code, '???')
        super().__init__()
    @property
    def name(self): return self._name
    @property
    def code(self): return self._code
    def __repr__(self):
        return f"DPLError(code={self._code!r}, name={self._name!r})"
    def __str__(self):
        return self.__repr__()

ERRORS = (
    "PREPROCESSING_ERROR",
    "SYNTAX_ERROR",
    "RUNTIME_ERROR",
    "PYTHON_ERROR",
    "PANIC_ERROR",
    "IMPORT_ERROR",
    "THREAD_ERROR",
    "TYPE_ERROR",
    "NAME_ERROR",
    "COMPAT_ERROR",
    "FILE_NOT_FOUND_ERROR"
)

META_ERR = None
PREPROCESSING_FLAGS = None

def error_setup_meta(scope):
    global META_ERR, PREPROCESSING_FLAGS
    META_ERR = scope
    PREPROCESSING_FLAGS = scope["preprocessing_flags"]
    scope["err"].update({"builtins": ERRORS, "defined_errors": list(ERRORS)})
    
    for pos, name in enumerate(ERRORS, 1):
        globals()[name] = pos
        META_ERR[name] = pos
        META_ERR[pos] = name

def register_error(name, value=None):
    if name in META_ERR:
        return META_ERR[name]
    META_ERR["defined_errors"].append(name)
    META_ERR[name] = (
        err_id := len(META_ERR["defined_errors"]) if value is None else value
    )
    META_ERR[err_id] = name
    ERRORS_DICT[err_id] = name
    return err_id

STOP_RESULT = -1
SKIP_RESULT = -2
FALLTHROUGH = -3
STOP_FUNCTION = -4

ERRORS_DICT = {
    globals().get(name): name for name in filter(lambda x: x.endswith("ERROR"), dir())
}

def my_print(*args, **kwargs):
    if PREPROCESSING_FLAGS["RUNTIME_ERRORS"]:
        print(*args, **kwargs)

def get_error_string(name, message):
    return None if name not in ERRORS_DICT else f"err:{ERRORS_DICT.get(name)}:{message}"

def pre_error(pos, file, cause=None):
    og_print(f"\n[Preprocessing Error]\nError in line {pos} file {file!r}")
    if cause is not None:
        og_print(f"Cause:\n{cause}")


def error(pos, file, cause=None):
    og_print(f"\nError in line {pos} file {file!r}")
    if cause is not None:
        og_print(f"Cause:\n{cause}")


og_print = my_print  # Use this to always call an error even when silent
is_silent = []

def info(text, show_date=True):
    if show_date:
        og_print(f"   [INFO] {datetime.datetime.now()}: {text}")
    else:
        og_print(f"   [INFO]: {text}")


def warnf(pos, file, text):
    og_print(f"\nWarning for line {pos} file {file!r}\n[WARNING]: {text}")


def warn(text, show_date=True):
    if show_date:
        og_print(f"[WARNING] {datetime.datetime.now()}: {text}")
    else:
        og_print(f"[WARNING]: {text}")

def pre_info(text, show_date=True):
    if show_date:
        og_print(f"   [INFO PRE] {datetime.datetime.now()}: {text}")
    else:
        og_print(f"   [INFO PRE]: {text}")


def pre_warn(text, show_date=True):
    if show_date:
        og_print(f"[WARNING PRE] {datetime.datetime.now()}: {text}")
    else:
        og_print(f"[WARNING PRE]: {text}")

# make the errors toggleable
def silent():
    global print
    og_print = lambda *x, **y: ...
    is_silent.append(0)

def active():
    global print
    is_silent.pop()
    if not is_silent:
        og_print = my_print


import json
import re
import os
from . import info
from traceback import format_exc
from prompt_toolkit.lexers import Lexer
from prompt_toolkit.styles import Style

# Your JSON config as a string
try:
    json_config = open(os.path.join(info.BINDIR, "repl_conf/colors_n_stuff.json")).read()
    print("Loaded highlighter config...")
except Exception as e:
    json_config = '{"classes":{}}'
    print("Something went wrong with loading the highlighter config...\nSee 'log.txt' in the dpl directory.")
    with open(os.path.join(info.BINDIR, "log.txt"), "w") as f:
        f.write(format_exc())

# Parse the config
config = json.loads(json_config)
classes = config['classes']

# Extract styles and word lists
class_styles = {}
word_map = {}  # word -> class name
regex_matchers = []  # (class name, regex)
sw_map = {}  # word -> class name

for class_name, rule in classes.items():
    class_styles[class_name] = rule['style']
    if 'words' in rule:
        for word in rule['words']:
            word_map[word] = class_name
    if 'match' in rule:
        regex_matchers.append((class_name, re.compile(eval(str(repr(rule['match']))))))
    if 'startswith' in rule:
        sw_map[class_name] = rule["startswith"]

# Construct the prompt_toolkit Style object
style = Style.from_dict(class_styles)

# Custom Lexer
class DPLLexer(Lexer):
    def lex_document(self, document):
        lines = document.lines

        def get_line(i: int):
            text = lines[i]
            tokens = []
            
            for class_name, word in sw_map.items():
                if text.startswith(word):
                    tokens.append((f'class:{class_name}', text))
                    return tokens
            
            # Tokenize line by words
            i = 0
            while i < len(text):
#                if text[i] == '"':
#                    end = text[i+1:].find('"')
#                    if end == -1:
#                        string = text[i:]
#                    else:
#                        string = text[i:i+2+end]
#                    tokens.append(("class:string", string))
#                    
#                    i += len(string)
#                    if i >= len(text):
#                        break
                if text[i] == '"':
                    start = i
                    end = i + 1
                    tokens.append(("class:string", '"'))
                    while end < len(text):
                        char = text[end]
                        if char == '\\':
                            if end+1 < len(text): tokens.extend([
                                ("class:escape", "\\"),
                                ("class:escape", text[end+1])
                            ]); end += 2
                            else: tokens.extend([
                                ("class:escape", "\\"),
                            ]); end += 1
                            continue
                        end += 1
                        tokens.append(('class:string', char))
                        if char == '"':
                            break
                    i = end
                    if i >= len(text): break
                if text[i] == "'":
                    start = i
                    end = i + 1
                    tokens.append(("class:string", "'"))
                    while end < len(text):
                        char = text[end]
                        if char == '\\':
                            if end+1 < len(text): tokens.extend([
                                ("class:escape", "\\"),
                                ("class:escape", text[end+1])
                            ]); end += 2
                            else: tokens.extend([
                                ("class:escape", "\\"),
                            ]); end += 1
                            continue
                        if char == '$' and end + 1 < len(text) and text[end+1] == '{':
                            sub_end = end + 2
                            
                            continue
                        end += 1
                        tokens.append(('class:string', char))
                        if char == "'":
                            break
                    i = end
                    if i >= len(text): break

                if text[i].isspace():
                    tokens.append(('', text[i]))
                    i += 1
                    continue

                start = i
                while i < len(text) and not text[i].isspace():
                    i += 1
                word = text[start:i]

                class_name = word_map.get(word)
                if class_name:
                    tokens.append((f'class:{class_name}', word))
                else:
                    for class_name, regex in regex_matchers:
                        if (match:=regex.match(word)):
                            tokens.append((f"class:{class_name}", word))
                            break
                    else:
                        tokens.append(('', word))

            return tokens

        return get_line


import dill

def rset(dct, full_name, value, sep="."):
    "Set a variable"
    if not isinstance(full_name, str):
        return
    if "." not in full_name:
            dct[full_name] = value
    path = [*enumerate(full_name.split(sep), 1)][::-1]
    last = len(path)
    node = dct
    while path:
        pos, name = path.pop()
        if (
            pos != last
            and name in node
            and isinstance(node[name], dict)
        ):
            node = node[name]
        elif pos == last:
            node[name] = value

cache = {}

def config(code, pack=False):
    "Helper to create bindings for DPL"
    data = {}
    for lpos, line in enumerate(code.split("\n"), 1):
        line = line.lstrip()
        if line.startswith("#") or not line:
            continue
        elif line.startswith("scope:"):
            rset(data, line[6:].strip(), {})
            continue
        elif line.startswith("::"):
            module_name = line[2:].strip()
            if "." not in module_name:
                rset(data, module_name, mod:=__import__(module_name))
                continue
            module, *path = module_name.split(".")
            try:
                module = __import__(module)
            except Exception as e:
                raise e from None
            while path:
                name = path.pop(0)
                if hasattr(module, name) and len(path) > 1:
                    module = getattr(module, name)
                elif hasattr(module, name):
                    rset(data, name, getattr(module, name))
                else:
                    print(f"Error [line {lpos}]: Invalid attribute {name} from object {module} which has {dir(module)}!")
                    return None
            continue
        var_name, module_name = line.split("<=")
        module_name = module_name.strip()
        var_name = var_name.strip()
        if "." not in module_name:
            rset(data, var_name, mod:=__import__(module_name))
            continue
        module, *path = module_name.split(".")
        try:
            module = __import__(module)
        except Exception as e:
            raise e from None
        while path:
            name = path.pop(0)
            if hasattr(module, name) and len(path) > 1:
                module = getattr(module, name)
            elif hasattr(module, name):
                rset(data, var_name, getattr(module, name))
            else:
                print(f"Error [line {lpos}]: Invalid attribute {name} from object {module} which has {dir(module)}!")
                return None
    return data if not pack else dill.dumps(data)



if __name__ != "__dpl_require__":
    raise Exception("This must be included by a DuProL Extension script!")


def has_repr(obj):
    return "_internal" in obj and "_im_repr" in obj


def get_repr(func):
    frame = dpl.varproc.new_frame()
    dpl.varproc.nscope(frame)
    frame[-1]["_returns"] = ["result"]
    if func["self"] != dpl.state.bstate("nil"):
        frame[-1]["self"] = func["self"]
    err = dpl.run_code(func["body"], frame=frame)
    if err > 0:
        raise Exception(err)
    return frame[0].get("result", dpl.constants.none)



from io import TextIOWrapper

if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")


if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

helper = dpl.require("dpl_helpers/func_helper.py")
if isinstance(helper, Exception):
    raise helper
ext = dpl.extension(meta_name="io", alias=__alias__)
ext.items["output"] = modules.sys.stdout

dpl.type_checker.alias["TextIOWrapper"] = TextIOWrapper

@ext.add_func("print", "@ranged $$ :: any ...")
def myPrint(_, __, *args, end="", sep=" "):
    args = list(args)
    for pos, arg in enumerate(args):
        if isinstance(arg, dict) and helper.has_repr(arg):
            arg[pos] = helper.get_repr(arg["_im_repr"])
    print(*args, end=end, sep=sep, file=ext.items["output"], flush=True)

@ext.add_func(typed="@ranged $$ :: any ...")
def println(_, __, *args, sep=" "):
    args = list(args)
    for pos, arg in enumerate(args):
        if isinstance(arg, dict) and helper.has_repr(arg):
            args[pos] = helper.get_repr(arg["_im_repr"])
    print(*args, sep=sep, file=ext.items["output"], flush=True)


@ext.add_func(typed="@ranged $$ :: any ...")
def rawprint(_, __, *args, sep=" ", end=""):
    print(*args, sep=sep, file=ext.items["output"], end=end, flush=True)


@ext.add_func(typed="@ranged $$ :: any ...")
def rawprintln(_, __, *args, sep=" "):
    print(*args, sep=sep, file=ext.items["output"], flush=True)


@ext.add_func("input", "%$$[str]\n%$$[str,str]")
def myInput(frame, __, prompt=None, name=None):
    res = input(prompt if prompt not in dpl.falsy else "")
    if name is not None:
        dpl.varproc.rset(frame[-1], name, res)


@ext.add_func(typed="$$ :: TextIOWrapper")
def setoutputfile(_, __, file):
    ext.items["output"] = file


@ext.add_func(typed="@ranged $$ :: any ...")
def rawoutput(_, __, *values):
    s = []
    for i in values:
        if isinstance(i, int):
            try:
                s.append(chr(i))
            except:
                s.append(str(i))
        elif isinstance(i, bytes):
            s.append(i.decode("utf-8"))
        else:
            s.append(str(i))
    modules.sys.stdout.write("".join(s))
    modules.sys.stdout.flush()


@ext.add_func(typed="%$$")
def flush(_, __):
    modules.sys.stdout.flush()



if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

ext = dpl.extension(meta_name="compiler")
# supported targets
ext.items["targets"] = ["python"]
# compilation result
ext.items["result"] = []
# curent indent
ext.items["indent"] = 0
# curent target (default python)
ext.items["target"] = "python"

compile_error = dpl.error.register_error("COMPILE_ERROR")

def indent(code):
    return (ext.items["indent"]*"  ")+code.replace("\n", "\n"+(ext.items["indent"]*"  "))

def process_arg(arg):
    if ext.items["target"] == "python":
        return str(arg)
    else:
        raise Exception(f"Unsupported target {ext.items['target']}")

def process_args(args):
    return ", ".join(map(process_arg, args))

@ext.add_func()
def compile_body(_, __, main_body):
    for [pos, file, ins, args] in main_body:
        match [ins, *args]:
            case ["printf", *args]:
                if ext.items["target"] == "python":
                    ext.items["result"].append(indent(f"print({process_args(args)})"))
            case ["set", var, "=", value]:
                if ext.items["target"] == "python":
                    ext.items["result"].append(indent(f"{var} = {process_arg(value)}"))
            case _:
                return f"err:{compile_error}:Couldnt compile {ins!r}"

@ext.add_func()
def collect(frame, _, name):
    frame[-1][name] = "\n".join(ext.items["result"])
    ext.items["result"].clear()